#include <new>
#include <memory.h>

#include <SDL_endian.h>

#include "sound.h"
#include "error.h"


#if SDL_BYTEORDER == SDL_BIG_ENDIAN
#define	AUDIOFORMAT	AUDIO_S16MSB	// 16ӥå椢
#else
#define	AUDIOFORMAT	AUDIO_S16		// 16ӥå椢
#endif


////////////////////////////////////////////////////////////////
// 󥹥ȥ饯
////////////////////////////////////////////////////////////////
cRing::cRing( void )
{
	Buffer = NULL;
	Size = Wpt = Rpt = Num = 0;
}


////////////////////////////////////////////////////////////////
// ǥȥ饯
////////////////////////////////////////////////////////////////
cRing::~cRing( void )
{
	if( Buffer ) delete[] Buffer;
}


////////////////////////////////////////////////////////////////
// Хåե
//
// :	size		ץ
// :	BOOL		TRUE: FALSE:
////////////////////////////////////////////////////////////////
BOOL cRing::InitBuffer( int size )
{
	PRINTD1( SND_LOG, "[cRing][Init] Size:%d\n", size );
	
	if( Buffer ) delete [] Buffer;
	
	Size = size*MULTI;	// ºݤΥХåեMULTI
	Wpt = Rpt = Num = 0;
	
	try{
		Buffer = new int[Size];
		memset( Buffer, 0, Size );
	}
	// new ˼Ԥ
	catch( std::bad_alloc ){
		Error::SetError( Error::MemAllocFailed );
		return FALSE;
	}
	
	return TRUE;
}


////////////////////////////////////////////////////////////////
// ɹ
//
// :	idx		ǥå(-1:1ǡɤǥݥ󥿿ʤ)
//								(0-:idx1ǡɤ ݥ󥿿ʤʤ)
// :	int		ǡ
////////////////////////////////////////////////////////////////
int cRing::Get( int idx )
{
	if( Num ){
		int data = 0;
		SCritical::Lock();
		if( idx < 0 ){
			data = Buffer[Rpt++];
			if( Rpt == Size ) Rpt = 0;
			Num--;
		}else{
			if( idx < Size ){
				int pt = Rpt + idx;
				if( pt >= Size ) pt -= Size;
				data = Buffer[pt];
			}
		}
		SCritical::UnLock();
		return data;
	}else{
		return 0;
	}
}


////////////////////////////////////////////////////////////////
// 
//
// :	data		ǡ
// :	BOOL		TRUE: FALSE:(ХåեäѤ)
////////////////////////////////////////////////////////////////
BOOL cRing::Put( int data )
{
	if( Num < Size ){
		SCritical::Lock();
		Buffer[Wpt++] = data;
		if( Wpt == Size ) Wpt = 0;
		Num++;
		SCritical::UnLock();
		return TRUE;
	}else
		return FALSE;
}


////////////////////////////////////////////////////////////////
// ̤ɥǡ
//
// :	ʤ
// :	int			̤ɥץ
////////////////////////////////////////////////////////////////
int cRing::ReadySize( void )
{
	return Num;
}


////////////////////////////////////////////////////////////////
// ĤХåե
//
// :	ʤ
// :	int			ĤХåեץ
////////////////////////////////////////////////////////////////
int cRing::FreeSize( void )
{
	return Size - Num;
}


////////////////////////////////////////////////////////////////
// Хåե
//
// :	ʤ
// :	int			Хåեץ
////////////////////////////////////////////////////////////////
int cRing::GetSize( void )
{
	return Size;
}




////////////////////////////////////////////////////////////////
// 󥹥ȥ饯
////////////////////////////////////////////////////////////////
SndDev::SndDev( void )
{
	Volume = 0;
}


////////////////////////////////////////////////////////////////
// ǥȥ饯
////////////////////////////////////////////////////////////////
SndDev::~SndDev( void ){}


////////////////////////////////////////////////////////////////
// 
//
// :	vol			(0-100)
// :	ʤ
////////////////////////////////////////////////////////////////
void SndDev::SetVolume( int vol )
{
	Volume = min( max( vol, 0 ), 100 );
}


////////////////////////////////////////////////////////////////
// ȥ꡼๹
//
// :	samples		륵ץ
// :	int			ץ
////////////////////////////////////////////////////////////////
int SndDev::SoundUpdate( int samples )
{
	return 0;
}


////////////////////////////////////////////////////////////////
// 󥹥ȥ饯
////////////////////////////////////////////////////////////////
SND6::SND6( void )
{
	memset( &ASpec, 0, sizeof(SDL_AudioSpec) );
	for( int i=0; i<MAXSTREAM; i++ ) RB[i] = NULL;
	Volume = 0;
}


////////////////////////////////////////////////////////////////
// ǥȥ饯
////////////////////////////////////////////////////////////////
SND6::~SND6( void )
{
	Pause();
	// ǥǥХĤ
	SDL_CloseAudio();
}


////////////////////////////////////////////////////////////////
// 
//
// :	p6			ʬȤؤΥ֥ȥݥ
//			callback	ХåؿؤΥݥ
//			rate		ץ󥰥졼
//			size		Хåե(Ψ)
// :	BOOL		TRUE: FALSE:
////////////////////////////////////////////////////////////////
BOOL SND6::Init( void *p6, void (*callback)(void *, Uint8 *, int ), int rate, int size )
{
	PRINTD( SND_LOG, "[SND6][Init]\n" );
	
	ASpec.freq     = rate;							// ץ󥰥졼
	ASpec.format   = AUDIOFORMAT;					// եޥå
	ASpec.channels = 1;								// Υ
	ASpec.samples  = rate * size / VSYNC_HZ;		// Хåե(ץ)
	ASpec.callback = callback;						// Хåؿλ
	ASpec.userdata = p6;							// Хåؿ˼ʬȤΥ֥ȥݥ󥿤Ϥ
	
	for( int i=0; i<MAXSTREAM; i++ ) RB[i] = NULL;
	
	Volume = DEFAULT_MASTERVOL;
	
	// Хåե
	if( !cRing::InitBuffer( ASpec.samples ) ) return FALSE;
	
	// ǥǥХ򳫤
	if( SDL_OpenAudio( &ASpec, NULL ) < 0 ) return FALSE;
	
	PRINTD1( SND_LOG, " SampleRate : %d\n",     ASpec.freq );
	PRINTD1( SND_LOG, " Format     : 0x%04X\n", ASpec.format );
	PRINTD1( SND_LOG, " Channels   : %d\n",     ASpec.channels );
	PRINTD1( SND_LOG, " Size       : %d\n",     ASpec.size );
	PRINTD1( SND_LOG, " Silence    : %d\n",     ASpec.silence );
	
	return TRUE;
}


////////////////////////////////////////////////////////////////
// ȥ꡼³
//
// :	cRing *		Хåեݥ
// :	BOOL		TRUE: FALSE:
////////////////////////////////////////////////////////////////
BOOL SND6::ConnectStream( cRing *buf )
{
	PRINTD( SND_LOG, "[SND6][ConnectStream]\n" );
	
	if( !buf ) return FALSE;
	
	for( int i=0; i<MAXSTREAM; i++ ){
		if( !RB[i] ){
			// Хåե
			if( !buf->InitBuffer( ASpec.samples ) ) return FALSE;
			RB[i] = buf;
			return TRUE;
		}
	}
	
	return FALSE;
}


////////////////////////////////////////////////////////////////
// 
//
// :	ʤ
// :	ʤ
////////////////////////////////////////////////////////////////
void SND6::Play( void )
{
	PRINTD( SND_LOG, "[SND6][Play]\n" );
	
	SDL_PauseAudio( 0 );
}


////////////////////////////////////////////////////////////////
// 
//
// :	ʤ
// :	ʤ
////////////////////////////////////////////////////////////////
void SND6::Pause( void )
{
	PRINTD( SND_LOG, "[SND6][Pause]\n" );
	
	SDL_PauseAudio( 1 );
}


////////////////////////////////////////////////////////////////
// ץ󥰥졼ȼ
//
// :	ʤ
// :	int		ץ󥰥졼
////////////////////////////////////////////////////////////////
int SND6::GetSampleRate( void )
{
	return ASpec.freq;
}


////////////////////////////////////////////////////////////////
// ޥ
//
// :	vol		(0-100)
// :	ʤ
////////////////////////////////////////////////////////////////
void SND6::SetVolume( int vol )
{
	Volume = min( max( vol, 0 ), 100 );
}


////////////////////////////////////////////////////////////////
// ɻؿ
//
// :	samples		ץ
//			exbuf		Хåեݥ
// :	int			ץ
////////////////////////////////////////////////////////////////
int SND6::PreUpdate( int samples, cRing *exbuf )
{
	int exsam = 0;
	
	PRINTD3( SND_LOG,"PreUpdate %d %d %d \n", RB[0]->ReadySize(), RB[1]->ReadySize(), RB[2]->ReadySize() );
	
	for( int i=0; i<MAXSTREAM; i++ )
		if( RB[i] ) exsam = min( max( exsam, RB[i]->ReadySize() ), samples );
	
	for( int i=0; i<exsam; i++ ){
		int dat = 0;
		for( int j=0; j<MAXSTREAM; j++ )
			if( RB[j] ) dat += RB[j]->Get();
		
		dat = ( dat * Volume ) / 100;
		dat = min( max( dat, -32768 ), 32767 );
		
		cRing::Put( (Sint16)dat );
		// Хåե¸ߤн
		if( exbuf ) exbuf->Put( (Sint16)dat );
	}
	
	return exsam;
}


////////////////////////////////////////////////////////////////
// ɹؿ
//
// :	stream		ȥ꡼ߥХåեؤΥݥ
//			samples		ץ
// :	ʤ
////////////////////////////////////////////////////////////////
void SND6::Update( Uint8 *stream, int samples )
{
	Sint16 *str = (Sint16 *)stream;
	
	PRINTD3( SND_LOG, "[SND6][Update] Stream:%08X Samples:%d / %d\n", (int)stream, samples, cRing::ReadySize() );
	
	for( int i=0; i<samples; i++ ){
		*(str++) = (Sint16)cRing::Get();
	}
}
